/* $Header:   P:/source/cwright/cwstart/m.c_v   1.33   15 Nov 1994 15:08:16   ericj  $ */
/*
 *
 *                Copyright 1991,1992,1993 Starbase Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Starbase Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Starbase Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Starbase Corp.
 *
 *    Owner:
 *        Starbase Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 * Matlab Implementation written by Charles H. Frankel, frankel@cs.uchicago.edu
 * The Unversity of Chicago Department of Computer Science
 * This module was written for MS Visual C++ Version 2.0
 * October, 1995
 *
 */

//#include "win_filt.h"
#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"


#define MAX_BUF_LEN    100

static LPVOID _m_hash_names        = NIL;  /* pointer to the hash table containing */
					   /* all 'ASM' keyword names. */
static LPVOID _m_hash_names_1      = NIL;  /* pointer to the hash table containing */
					   /* user defined keyword names. */
static int _m_hash_size_1          = 47;
static LPVOID _m_hash_names_2      = NIL;  /* pointer to the hash table containing */
					   /* user defined keyword names. */
static int _m_hash_size_2          = 47;

static LPVOID _m_hash_templates    = NIL;  /* pointer to the hash table containing */
					   /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _m_hash_templates.
 */
typedef struct _LanguageTemplate {         
	LPSTR abbrev;
	LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR ASM_INDENT_STR = "\377";

/*
 * _m_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _m_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _m_templates[] = {
	{ "cProc",   "cProc &, <PUBLIC>,<ds,es,si,si>\ncBegin\ncEnd\n" },
	{ NIL,       NIL },
};

/* Define the list of 'm' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
/* Mostly keywords, but a few -very- commonly used function calls as well
   CHF */
LPSTR _m_names[] = {
	"break",        "case",         "if",       "zeros",
	"continue",     "default",  	"do",       "else",
	"for",          "switch",   	"eval",     "nargin",
	"round",        "ceil",         "floor",    "elseif",	"return",
	"while",         "end",          "size",    "function", "clear",
    NIL
};

static long _m_in_comment( long offset );

void DLL _m_init( void );
BOOL DLL _m_expand( void );
void DLL _m_language_colors_update( long firstline, long lastline );
void DLL _m_language_colors( long numlines );
void DLL _m_assign_template( LPSTR abbrev, LPSTR template );
void DLL _m_add_keyword( LPSTR keyword, int table );
BOOL DLL _m_indent( void );

LIBMAIN    /* defines the LibMain() function for Borland C */

void DLL 
_init()
{
	_m_init();
}

/*
 ** _m_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_m_init( void )
{
	static BOOL firstTime = TRUE;
	int    i;

	if (firstTime)
	{
		LibExport( "BOOL _m_expand");
		LibExport( "_m_assign_template LPSTR LPSTR");
		LibExport( "_m_add_keyword LPSTR int");
		LibExport( "BOOL _m_indent");
		LibExport( "_m_language_colors_update long long");
		LibExport( "_m_language_colors long");
		LibExport( "_m_hash_size int int" );
		LibExport( "_m_routines");

		firstTime = FALSE;

		/* Create a hash table to hold all of the 'ASM' keyword names.
		 */
		_m_hash_names = HashCreateTable( 539 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;

		while ( _m_names[ i ] != NIL )
		{
			LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _m_hash_names, 
					_m_names[i], strlen( _m_names[i] ) );
			if (lpf)
			{
				*lpf = StrNew( _m_names[i] );
				i++;
			}
		}

		/* Create a hash table to hold all of the 'ASM' template names.
		 */
		_m_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;

		while ( _m_templates[ i ].abbrev != NIL )
		{
			_m_assign_template( _m_templates[i].abbrev, _m_templates[i].template );
			i++;
		}
	}
}

void DLL
_m_hash_size( int ht, int size )
{
	if (size > 0)
	{
		if (ht == 1)
			_m_hash_size_1 = size;
		else if (ht == 2)
			_m_hash_size_2 = size;
	}
}


/*
 ** _m_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'ASM' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_m_expand( void )
{
	LPVOID htEntry;
	LPSTR  _m_string = NULL;
	LPSTR  str;
	int    i = 0;

	MarkSavePos();

	/* Check to see if we are currently positioned inside a 'ASM' comment
	 */
	if (_m_in_comment( -1L ) != -1)
		goto _m_expand_incomplete;

	str = BufReadStr( 65535 );
	while (str[i] && !isspace(str[i]))
		i++;
	str[i] = '\0';

	_m_string = StrTrim( str, " \t" );
	StrFree( str );

	if (!_m_string)
		goto _m_expand_incomplete;

	/* Search for a keyword match in the hash table.
	 */
	htEntry = HashFindEntry( _m_hash_templates, _m_string, strlen( _m_string ) );
	StrFree( _m_string );

	if (htEntry)
	{
		/* A template exists for the current keyword, delete the keyword
		 * and insert the template.
		 */
		MarkDropPos( 1 );
		str = *(LPSTR*)htEntry;
		BufDelToEOL();

		_ext_expand_template( str );

	} 
	else
		goto _m_expand_incomplete;
	return TRUE;

_m_expand_incomplete:
	MarkRestorePos();

	return FALSE;
}

/*
 ** _m_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
void DLL
_m_language_colors_update( long firstline, long lastline )
{
	long startoff;

	if (firstline && lastline)
	{
		/* Begin coloring at the previous comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		if (firstline == 1)
			_PosPrevLine( 0L );
		else
			_PosNextLine( firstline - 1);
		startoff = _PosQOffset();

		/* Clear all attributes for the line range and
		 * the reset them within _m_language_colors.
		 */
		AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
		_PosInit( startoff );
		_m_language_colors( lastline - firstline + 1 );
	}
}

/*
 ** _m_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_m_language_colors( long numlines )
{
	long  startoff;
	long  startline;
	long  endoff;
	long  inc;
	WORD  ch;
	LPSTR buf;
	int   i = 0;
	int   comment_color = ColorComments( -1 );
	int   keyword_color = ColorKeywords( -1 );
	int   quote_color      = ColorAlternate1( -1 );
	int   number_color     = ColorAlternate2( -1 );

	buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

	/* Start at the top of the buffer or current position 
	 */
	if (numlines == -1)
	{
		_PosInit( 0 );  
		numlines = 0x7FFFFFFF;
	}
	else
	{
		startoff  = _PosQOffset();
		startline = _PosQLine();
		endoff    = _m_in_comment( startoff );
		if (endoff == -1)
			_PosInit( startoff );
		else
		{
			_PosInit( endoff );
			numlines += (startline - _PosQLine());
		}
	}
	ch = _PosCurrentChar();
	startoff = _PosQOffset();


	while ((ch != EOF_CHAR) && (numlines > 0))
	{
		if (ch == '\n')
			numlines--;

		if (ch == '"' || ch == '\'')
		{
			UINT quote = ch;

	    startoff = _PosQOffset();

			do 
			{
				ch = _PosNextChar();
				if (ch == '\\')
				{
					ch = _PosNextChar();
					if ((ch == quote) || (ch == '\\'))
						ch = _PosNextChar();
				}
				if ((ch == EOF_CHAR) || (ch == '\n'))
					break;
			} while (ch != quote);

			if (ch == quote)
				ch = _PosNextChar();

			/* color the string */
			if (quote_color)
			{
				endoff = _PosQOffset();
			_PosInit( startoff );
				_PosSetColor( quote_color, endoff - startoff );
				_PosInit( endoff );
			}
			continue;
		}
		else if (number_color && isdigit( ch ))
		{
	    int j = 0;
			UINT lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			startoff = _PosQOffset();
			do
			{
		j++;
				ch = _PosNextChar();
			} while (isxdigit( ch ) || (ch == '.'));   // || (ch == 'x' && j == 1));

			if ((  lastch != ' ' 
				&& lastch != '\t' 
				&& lastch != '\r' 
				&& lastch != '\n' 
				&& lastch != EOF_CHAR
				&& !ispunct( lastch ))
				|| lastch == '_'
				|| (ch == '_'))
				continue;

			if (strchr( "bBdDhHoOtTyY", ch ) != NULL)
		ch = _PosNextChar();

			endoff = _PosQOffset();
		_PosInit( startoff );
			_PosSetColor( number_color, endoff - startoff );
			_PosInit( endoff );

		}
		else if (isalpha( (BYTE)ch ) || (ch == '_') || (ch == '.'))
		{
			WORD lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			i = 0;
			startoff = _PosQOffset();

			do
			{
				if (i < MAX_BUF_LEN)
					buf[i++] = (char)ch;
				ch = _PosNextChar();
			} while (isalnum( (BYTE)ch ) || (ch == '_'));

			if (strchr( " \t\r\n[(:", lastch ) == NULL)
				i = 0;

			continue;
		}
		else if (i)
		{
			if (isalpha( (BYTE)buf[0] ))
			{
				int color = 0;
				if (_m_hash_names_1 && HashFindEntry( _m_hash_names_1, buf, i ))
					color = ColorAlternate5( -1 );
				else if (_m_hash_names_2 && HashFindEntry( _m_hash_names_2, buf, i ))
					color = ColorAlternate6( -1 );
				else if (HashFindEntry( _m_hash_names, buf, i ))
					color = keyword_color;
	
				if (color)
				{
					endoff = _PosQOffset();
					_PosInit( startoff );
					_PosSetColor( color, endoff - startoff );
					_PosInit( endoff );
				}
			}
			i = 0;
		} 

		if (ch == '%')
		{
			startoff = _PosQOffset();

			inc = _PosNextLine( 1 );
			numlines--;
			endoff = _PosQOffset();
			_PosInit( startoff );
			_PosSetColor( comment_color, endoff - startoff - inc );
			_PosInit( endoff );
			ch = _PosCurrentChar();
			continue;
		}
		ch = _PosNextChar();
	}
	MemFree( buf );
}


void DLL
_m_add_keyword( LPSTR keyword, int table )
{
	if (keyword && *keyword && table >= 0)
	{
		LPSTR XFAR *lpf;
		LPVOID  ht;

		switch( table )
		{
		case 0:
			ht = _m_hash_names;
			break;
		case 1:
			if (_m_hash_names_1 == NULL)
				_m_hash_names_1 = HashCreateTable( _m_hash_size_1, sizeof( LPSTR ), 0 );
			ht = _m_hash_names_1;
			break;
		case 2:
			if (_m_hash_names_2 == NULL)
				_m_hash_names_2 = HashCreateTable( _m_hash_size_2, sizeof( LPSTR ), 0 );
			ht = _m_hash_names_2;
			break;
		}

		lpf = (LPSTR XFAR *)HashGetEntry( ht, 
								keyword, strlen( keyword ) );
		if (lpf)
			*lpf = StrNew( keyword );
	}
}



/*
 ** _m_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'ASM' template strings
 *    without changing the code.
 */
void DLL
_m_assign_template( LPSTR abbrev, LPSTR template )
{
	_ext_assign_template( _m_hash_templates, abbrev, template );
}


/*
 ** _m_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'ASM' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in ASM_INDENT_STR - perform smart
 *       indentation.  
 */
BOOL DLL
_m_indent( void )
{
	long   curline;
	long   col;
	LPSTR  _m_string;
	LPSTR  str;
	long   matchpos;

	MarkSavePos();

	curline = BufQCurrentLine();

	/* Check to see if we are currently positioned inside a 'ASM' comment
	 */
	if (_m_in_comment( -1L ) != -1)
		goto _m_indent_incomplete;

	/* Not in a comment so indent according to the current 'ASM' construct.
	 */
	if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
			|| (BufQCurrentLine() != curline) )
		goto _m_indent_incomplete;

	/* Read the keyword at the beginning of the current line.
	 */
	col = BufQCurrentCol();
	SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
	str = BufReadStr( col - BufQCurrentCol() );
	_m_string = StrTrim( str, " \t" );
	StrFree( str );

	/* Set prefix and suffix to \377 to perform a quick lookup of the name
	 */
	str = StrNew( "\377" );
	_m_string = StrAppend( _m_string, str );
	str = StrAppend( str, _m_string );
	StrFree( _m_string );
	
	matchpos = StrMatch( str, ASM_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );
	StrFree( str );

	if (!matchpos)
		/* Not a 'ASM' keyword, do normal indentation. */
		goto _m_indent_incomplete;
		
	MarkRestorePos();

	_ext_insert_indented_EOL();
	BufInsertChar( '\t' );
	return TRUE;

	_m_indent_incomplete:
		MarkRestorePos();

	return FALSE;
}



/*--  Private functions --*/


/*
 ** _m_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'ASM' comment by search backward for a // or /* 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_m_in_comment( long offset )
{
	WORD  ch;

	/* Check to see if we are currently positioned inside a 'ASM' comment
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	do
	{
		if (ch == '\n')
			return -1;
		else if (ch == '%')
			return _PosQOffset();  /* currrent line is in a comment */
		ch = _PosPrevChar();
	} while (ch != EOF_CHAR);

	return -1L;
}

void DLL
_m_routines( void )
{
	int   matches = 0;
	LPSTR str     = "^[a-zA-Z0-9_].*\\((.|\n)*\\)";
	LPSTR stre    = "^[a-zA-Z0-9_].*\\((.|\n)*\\)[ \t\n]*\\c[^ \t\n]";
	long  curline;
	UINT  ch;
	long  lines;

#ifdef LIGHT
	if (Grep( "^[a-zA-Z0-9_].*\\)[ \t]*{?[ \t]*$" ))
		MsgMessage( "" );
	else
		MsgMessage( "No routines found." );
#else
	SysWaitStart();

	MarkSavePos();

	BufSetAscii();
	MovTopBuf();

	while (SrchFind( str, SEARCH_REGEX | SEARCH_FORWARD, 0 ))
	{
		curline = BufQCurrentLine();

		SrchFind( stre, SEARCH_REGEX | SEARCH_FORWARD, 0 );

		ch   = BufReadChar();
		if (ch != ';')
		{
			if (!matches)
				AttrSetVisible( 0, 0, 0 );
			matches++;
			lines = BufQCurrentLine() - curline
						- !!(ch == '{');
			AttrSetVisible( curline, curline + (lines < 0L ? 0L : lines), 1 );
			MovHome();
		}

		if (!MovDown( 1 ))
			break; 
	}

	if (matches)
	{
		BufSetCompact();
		MarkDropPos( 1 );
		MovTopBuf();
	}
	else
		MarkRestorePos();

	SysWaitEnd();

	MsgMessage( matches ? "" : "No routines found." );
#endif
}
